--  Original file:  http://weitz.de/files/einstein2.cpp
--  (c) Klaus Betzler 20011218 <Klaus.Betzler@uos.de>
--
--  Translated from C++ to Euphoria
--  by Juergen Luethje 2004-01-26 <support {AT} luethje {DOT} eu>
--  Permission for translating and redistributing the program kindly
--  granted by Prof. Klaus Betzler.
--
--  'Einstein's Riddle', the rules:
--
--  1 The Brit lives in the red house
--  2 The Swede keeps dogs as pets
--  3 The Dane drinks tea
--  4 The green house is on the left of the white house
--  5 The green house's owner drinks coffee
--  6 The person who smokes Pall Mall rears birds
--  7 The owner of the yellow house smokes Dunhill
--  8 The man living in the centre house drinks milk
--  9 The Norwegian lives in the first house
-- 10 The person who smokes Marlboro lives next to the one who keeps cats
-- 11 The person who keeps horses lives next to the person who smokes Dunhill
-- 12 The person who smokes Winfield drinks beer
-- 13 The German smokes Rothmans
-- 14 The Norwegian lives next to the blue house
-- 15 The person who smokes Marlboro has a neigbor who drinks water

without type_check
include cputime.e          -- Because the program is extremely fast,
                           -- we must use an improved time() function.
type unsigned (object x)
   if integer(x) then
      return x >= 0
   end if
   return 0
end type

function bit (unsigned n)
   return power(2,n)
end function

function shift_right (unsigned v, unsigned count)
   return v/power(2,count)
end function


constant
   false = 0,
   true = 1

constant
   yellow = bit(0),
   blue   = bit(1),
   red    = bit(2),
   green  = bit(3),
   white  = bit(4),

   norwegian = bit(5),
   dane      = bit(6),
   brit      = bit(7),
   german    = bit(8),
   swede     = bit(9),

   water  = bit(10),
   tea    = bit(11),
   milk   = bit(12),
   coffee = bit(13),
   beer   = bit(14),

   dunhill  = bit(15),
   marlboro = bit(16),
   pallmall = bit(17),
   rothmans = bit(18),
   winfield = bit(19),

   cat   = bit(20),
   horse = bit(21),
   bird  = bit(22),
   fish  = bit(23),
   dog   = bit(24)

constant Label = {
   "Yellow","Blue","Red","Green","White",
   "Norwegian","Dane","Brit","German","Swede",
   "Water","Tea","Milk","Coffee","Beer",
   "Dunhill","Marlboro","Pallmall","Rothmans","Winfield",
   "Cat","Horse","Bird","Fish","Dog"
}

constant
   color   = yellow+blue+red+green+white,
   country = norwegian+dane+brit+german+swede,
   drink   = water+tea+milk+coffee+beer,
   cigar   = dunhill+marlboro+pallmall+rothmans+winfield,
   animal  = cat+horse+bird+fish+dog

constant comb = {            -- simple rules
   brit+red,                 -- 1
   swede+dog,                -- 2
   dane+tea,                 -- 3
   green+coffee,             -- 5
   pallmall+bird,            -- 6
   yellow+dunhill,           -- 7
   winfield+beer,            -- 12
   german+rothmans           -- 13
}

constant combmask = {        -- corresponding selection masks
   country+color,
   country+animal,
   country+drink,
   color+drink,
   cigar+animal,
   color+cigar,
   cigar+drink,
   country+cigar
}

sequence house, result             -- These are the only variables!
house = {norwegian,blue,milk,0,0}  -- rules 8,9,14
result = repeat(0,5)

function SimpleRule (unsigned nr, unsigned which)
   if which < 9 then
      if and_bits(house[nr], combmask[which]) > 0 then
         return false
      else
         house[nr] = or_bits(house[nr], comb[which])
         return true
      end if
   else            -- rule 4
      if (nr=5) or (and_bits(house[nr],green)=0) then
         return false
      else
         if and_bits(house[nr+1], color) > 0 then
            return false
         else
            house[nr+1] = or_bits(house[nr+1], white)
            return true
         end if
      end if
   end if
end function

procedure RemoveSimple (unsigned nr, unsigned which)
   if which < 9 then
      house[nr] = and_bits(house[nr], not_bits(comb[which]))
   else
      house[nr+1] = and_bits(house[nr+1], not_bits(white))
   end if
end procedure

function DunhillRule (unsigned nr, integer side)  -- 11
   if ((side=1) and (nr=5))
   or ((side=-1) and (nr=1))
   or (and_bits(house[nr], dunhill)=0) then
      return false
   end if
   if and_bits(house[nr+side], animal) > 0 then
      return false
   end if
   house[nr+side] = or_bits(house[nr+side], horse)
   return true
end function

procedure RemoveDunhill (unsigned nr, unsigned side)
   house[nr+side] = and_bits(house[nr+side], not_bits(horse))
end procedure

function MarlboroRule (unsigned nr)    -- 10 + 15
   if and_bits(house[nr], cigar) > 0 then
      return false
   end if
   house[nr] = or_bits(house[nr], marlboro)

   if nr = 1 then
      if and_bits(house[2], animal+drink) > 0 then
         return false
      else
         house[2] = or_bits(house[2], cat+water)
         return true
      end if
   end if
   if nr = 5 then
      if and_bits(house[4], animal+drink) > 0 then
         return false
      else
         house[4] = or_bits(house[4], cat+water)
         return true
      end if
   end if

   for i = -1 to 1 by 2 do
      if and_bits(house[nr+i], animal) = 0 then
         house[nr+i] = or_bits(house[nr+i], cat)
         for k = -1 to 1 by 2 do
            if and_bits(house[nr+k], drink) = 0 then
               house[nr+k] = or_bits(house[nr+k], water)
               return true
            end if
         end for
      end if
   end for
   return false
end function

procedure RemoveMarlboro (unsigned m)
   house[m] = and_bits(house[m], not_bits(marlboro))
   if m > 1 then
      house[m-1] = and_bits(house[m-1], not_bits(cat+water))
   end if
   if m < 5 then
      house[m+1] = and_bits(house[m+1], not_bits(cat+water))
   end if
end procedure

procedure Recurse (unsigned recdepth)
   for n = 1 to 5 do
      if recdepth < 10 then          -- simple rules
         if SimpleRule(n, recdepth) then
            Recurse(recdepth+1)
            RemoveSimple(n, recdepth)
         end if
      else                           -- Dunhill and Marlboro
         for side = -1 to 1 by 2 do
            if DunhillRule(n, side) then
               for m = 1 to 5 do
                  if MarlboroRule(m) then
                     for r = 1 to 5 do
                        result[r] = house[r]
                     end for
                  else
                     RemoveMarlboro(m)
                  end if
               end for
               RemoveDunhill(n, side)
            end if
         end for
      end if
   end for
end procedure

function main()
   atom t

   t = time()
   Recurse(1)
   t = (time()-t)*1000000
   printf(1, "Computation Time: %.0f microsec\n\n", {t})

   if result[1] = 0 then
      puts(1, "No solution found !?!\n")
      return 1
   end if

   for i = 1 to 5 do
      if and_bits(result[i], animal) = 0 then
         for index = 0 to 24 do
            if shift_right(and_bits(result[i],country), index) = 1 then
               printf(1, "Fish Owner is the %s !!!\n\n", {Label[index+1]})
            end if
         end for
      end if
   end for

   for i = 1 to 5 do
      printf(1, "%d: ", {i})
      for index = 0 to 24 do
         if and_bits(shift_right(result[i],index), 1) = 1 then
            printf(1, "%-12s", {Label[index+1]})
         end if
      end for
      puts(1, "\n\n")
   end for
   return 0
end function

integer ErrCode
ErrCode = main()
puts(1, "\nPress Enter ...")
if getc(0) then end if
abort(ErrCode)
